#include <Planner/ActionsChain.h>

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/join.hpp>

#include <IO/WriteBuffer.h>
#include <IO/WriteHelpers.h>
#include <IO/Operators.h>
#include <IO/WriteBufferFromString.h>

namespace DB
{

ActionsChainStep::ActionsChainStep(ActionsDAGPtr actions_, AvailableOutputColumnsStrategy available_output_columns_stategy_)
    : actions(std::move(actions_))
    , available_output_columns_strategy(available_output_columns_stategy_)
{
    initialize();
}

ActionsChainStep::ActionsChainStep(ActionsDAGPtr actions_,
    AvailableOutputColumnsStrategy available_output_columns_stategy_,
    ColumnsWithTypeAndName additional_output_columns_)
    : actions(std::move(actions_))
    , available_output_columns_strategy(available_output_columns_stategy_)
    , additional_output_columns(std::move(additional_output_columns_))
{
    initialize();
}


void ActionsChainStep::finalizeInputAndOutputColumns(const NameSet & child_input_columns)
{
    child_required_output_columns_names.clear();

    auto child_input_columns_copy = child_input_columns;

    std::unordered_set<std::string_view> output_nodes_names;
    output_nodes_names.reserve(actions->getOutputs().size());

    for (auto & output_node : actions->getOutputs())
        output_nodes_names.insert(output_node->result_name);

    for (const auto & node : actions->getNodes())
    {
        auto it = child_input_columns_copy.find(node.result_name);
        if (it == child_input_columns_copy.end())
            continue;

        child_input_columns_copy.erase(it);
        child_required_output_columns_names.insert(node.result_name);

        if (output_nodes_names.contains(node.result_name))
            continue;

        actions->getOutputs().push_back(&node);
        output_nodes_names.insert(node.result_name);
    }

    actions->removeUnusedActions();
    /// TODO: Analyzer fix ActionsDAG input and constant nodes with same name
    actions->projectInput();
    initialize();
}

void ActionsChainStep::dump(WriteBuffer & buffer) const
{
    buffer << "DAG" << '\n';
    buffer << actions->dumpDAG();

    if (!additional_output_columns.empty())
    {
        buffer << "Additional output columns " << additional_output_columns.size() << '\n';
        for (const auto & column : additional_output_columns)
            buffer << "Name " << column.name << " type " << column.type->getName() << '\n';
    }

    if (!child_required_output_columns_names.empty())
    {
        buffer << "Child required output columns " << boost::join(child_required_output_columns_names, ", ");
        buffer << '\n';
    }
}

String ActionsChainStep::dump() const
{
    WriteBufferFromOwnString buffer;
    dump(buffer);

    return buffer.str();
}

void ActionsChainStep::initialize()
{
    auto required_columns_names = actions->getRequiredColumnsNames();
    input_columns_names = NameSet(required_columns_names.begin(), required_columns_names.end());

    available_output_columns.clear();

    /// TODO: Analyzer fix ActionsDAG input and constant nodes with same name
    std::unordered_set<std::string_view> available_output_columns_names;

    if (available_output_columns_strategy == AvailableOutputColumnsStrategy::ALL_NODES)
    {
        for (const auto & node : actions->getNodes())
        {
            if (available_output_columns_names.contains(node.result_name))
                continue;

            available_output_columns.emplace_back(node.column, node.result_type, node.result_name);
            available_output_columns_names.insert(node.result_name);
        }
    }
    else if (available_output_columns_strategy == AvailableOutputColumnsStrategy::OUTPUT_NODES)
    {
        for (const auto & node : actions->getOutputs())
        {
            if (available_output_columns_names.contains(node->result_name))
                continue;

            available_output_columns.emplace_back(node->column, node->result_type, node->result_name);
            available_output_columns_names.insert(node->result_name);
        }
    }

    available_output_columns.insert(available_output_columns.end(), additional_output_columns.begin(), additional_output_columns.end());
}

void ActionsChain::finalize()
{
    if (steps.empty())
        return;

    /// For last chain step there are no columns required in child nodes
    NameSet empty_child_input_columns;
    steps.back().get()->finalizeInputAndOutputColumns(empty_child_input_columns);

    Int64 steps_last_index = steps.size() - 1;
    for (Int64 i = steps_last_index; i >= 1; --i)
    {
        auto & current_step = steps[i];
        auto & previous_step = steps[i - 1];

        previous_step->finalizeInputAndOutputColumns(current_step->getInputColumnNames());
    }
}

void ActionsChain::dump(WriteBuffer & buffer) const
{
    size_t steps_size = steps.size();

    for (size_t i = 0; i < steps_size; ++i)
    {
        const auto & step = steps[i];
        buffer << "Step " << i << '\n';
        step->dump(buffer);

        buffer << '\n';
    }
}

String ActionsChain::dump() const
{
    WriteBufferFromOwnString buffer;
    dump(buffer);

    return buffer.str();
}

}
